#include "embossStringView.h"
#include "frmHelper.h"

#define EMBOSS_OFFSET_X						1	// in pixels, must be non-negative
#define EMBOSS_OFFSET_Y						1	// in pixels, must be non-negative

#define BE_FONT_RASTERIZER_LEGACY_OFFSET	1

#define	COLOR_BOUNDING_BOX					ColorRGBA(184, 184, 255)

EmbossStringView::EmbossStringView(const BRect &bounds,
								   const char *name,
								   const char *text,
								   uint32 resizeFlags,
								   uint32 flags) :

	BView(bounds, name, resizeFlags, flags),

	mText(text),
	mAlign(B_ALIGN_LEFT)
{ }

EmbossStringView::~EmbossStringView()
{ }

void EmbossStringView::Draw(BRect bounds)
{
	if (mText.CountChars() == 0)
		return;

	BFont font;
	GetFont(&font);

	font_height	fontHeight;
	font.GetHeight(&fontHeight);

	BPoint pointPen(0, (Bounds().bottom - fontHeight.descent) + BE_FONT_RASTERIZER_LEGACY_OFFSET);

	const char *string = mText.String(); 
	BRect boundingBox;

	escapement_delta delta0;
	delta0.nonspace = 0;
	delta0.space = 0;

	font.GetBoundingBoxesForStrings(&string, 1, B_SCREEN_METRIC, &delta0, &boundingBox);

	boundingBox.right += EMBOSS_OFFSET_X;
	boundingBox.bottom += EMBOSS_OFFSET_Y;

	switch(mAlign)
	{
	case B_ALIGN_LEFT:

		pointPen.x = 0;
		break;

	case B_ALIGN_RIGHT:

		pointPen.x = Bounds().Width() - boundingBox.Width();
		break;

	case B_ALIGN_CENTER:

		pointPen.x = (Bounds().Width() - boundingBox.Width()) * .5;
		break;
	}

	BPoint pointBaseline(pointPen.x, pointPen.y - BE_FONT_RASTERIZER_LEGACY_OFFSET);
	BRect boundsText(pointBaseline.x, pointBaseline.y + boundingBox.top,
					 pointBaseline.x + boundingBox.Width(), pointBaseline.y + boundingBox.bottom);	

	if ((bounds & boundsText).IsValid())
	{
#ifdef __BOUNDING_BOX

		rgb_color oldColor = HighColor();
		SetHighColor(COLOR_BOUNDING_BOX);
		StrokeRect(boundsText);
		SetHighColor(oldColor);

#endif // __BOUNDING_BOX

		rgb_color highColor = HighColor();
		SetHighColor(::tint_color(::ui_color(B_MENU_ITEM_TEXT_COLOR), B_LIGHTEN_MAX_TINT));
		DrawString(mText.String(), BPoint(pointPen.x + EMBOSS_OFFSET_X, pointPen.y), &delta0);
		SetHighColor(highColor);
		DrawString(mText.String(), BPoint(pointPen.x, pointPen.y - EMBOSS_OFFSET_Y), &delta0);

		// passing any escapement delta to DrawString() supresses the left edge inset of
		// the first character. furthermore, the escapement delta itself gets discarded
		// when font spacing mode is B_BITMAP_SPACING and view is not sub-pixel precise.
	}
}

void EmbossStringView::ResizeToPreferred()
{
	float width, height;
	GetPreferredSize(&width, &height);

	ResizeTo(width, height);
} 

void EmbossStringView::GetPreferredSize(float *width, float *height)
{
	BFont		font;
	font_height fontHeight;

	GetFont(&font);
	font.GetHeight(&fontHeight);

	*height = fontHeight.ascent + fontHeight.descent + 1 + EMBOSS_OFFSET_Y;
	*width = font.StringWidth(mText.String());
}

void EmbossStringView::SetText(const char *text)
{
	mText = text;
}

const char * EmbossStringView::Text() const
{
	return mText.String();
}
 
void EmbossStringView::SetAlignment(alignment flag)
{
	if (flag != B_ALIGN_LEFT &&
		flag != B_ALIGN_RIGHT &&
		flag != B_ALIGN_CENTER)
		return;
		
	mAlign = flag;
}
 
alignment EmbossStringView::Alignment() const
{
	return mAlign;
}
