#ifndef __FRAMEWORK_DIRECT_WINDOW_H__
#define __FRAMEWORK_DIRECT_WINDOW_H__

#include <OS.h>
#include <Locker.h>
#include <DirectWindow.h>

#include "frmwrk.h"

// window ////////////////////////////////////////////////////////////////////////////

class FrmDirectWindow : public BDirectWindow, public Framework
{
	static				int32 FlippingThread(void *data);
	thread_id			mFlipThreadID;

	BLocker				mFlipLock;
	port_id				mFlipPort;

	struct BufferRecord
	{
		uint8 *			bits;
		sem_id			writeLock;		
		bool			unlockAtConnect;
		
		BufferRecord() :
			bits(NULL),
			writeLock(B_ERROR)
		{ }
	}
					  *	mRing;
	int32				mRingSize;
	int32				mDrawSelector;

	uint8 *				mFrontBits;
	uint8 *				mBackBits;

	struct BitsInfo
	{
		int32			bytesPerRow; 
		uint32			bitsPerPixel;
		color_space		pixelFormat;
	}
						mFrontBitsInfo, mBackBitsInfo;

	struct ClipInfo
	{
		clipping_rect	windowBounds;
		uint32			clipRectsCount;
		clipping_rect *	clipRects;
	}
						mFrontClipInfo;

	int32				mFrameWidth;
	int32				mFrameHeight;

	volatile bool		mConnected;

public:

	FrmDirectWindow(FrameworkParams &params, status_t &error, const char * &error_text);
	~FrmDirectWindow();
		
	virtual bool		QuitRequested();
	virtual void		DirectConnected(direct_buffer_info *info);

	// interface methods to the drawing thread
	virtual status_t	frame_open(Buffer &);
	virtual void	 	frame_close();
};

#endif // __FRAMEWORK_DIRECT_WINDOW_H__
