#ifndef __FRAMEWORK_H__
#define __FRAMEWORK_H__

#include <GraphicsDefs.h>
#include <Point.h>

class Framework
{
public:

	struct Buffer
	{
		int32			frameHeight;
		int32			frameWidth;
		color_space		pixelFormat;
		uint32			rowStride;
		void *			bits;
	};

	// interface to the drawing thread
	virtual status_t	frame_open(Buffer &) = 0;
	virtual void	 	frame_close() = 0;

	static const char *	__painter_name();
	static int32		__highest_supported_bitdepth();
};

inline const char * Framework::__painter_name()
{
	return "an_amazing_drawing_thread";
}

inline int32 Framework::__highest_supported_bitdepth()
{
	size_t pixel_chunk, pixels_per_chunk, row_alignment;

	::get_pixel_size_for(B_RGBA32, &pixel_chunk, &row_alignment, &pixels_per_chunk);

	return (pixel_chunk * 8) / pixels_per_chunk;
}

class FrameworkParams
{
public:

	uint32 				numRingBuffers;
	color_space			pixelFormat;
	BPoint				frameBounds;
};

#endif // __FRAMEWORK_H__
