#ifndef __MEMBLT_H__
#define __MEMBLT_H__

#include <memory.h>

#ifdef __INTEL__

inline void x86_memblt_long_aligned_src(const void *src, void *dst, uint32 len)
{
	asm volatile (
   "leal	(%%esi,%%esi,2),%%ecx
	andl	$3,%%ecx
	subl	%%ecx,%%eax
	jle		__EndBytes
rep
	movsb
	movl	%%eax,%%ecx
	andl	$3,%%eax
	shrl	$2,%%ecx
rep
	movsl
__EndBytes:
	addl	%%eax,%%ecx
rep
	movsb"

	: "=S" (src), "=D" (dst), "=a" (len)
	: "0" (src), "1" (dst), "2" (len)
	: "cc", "%ecx" );
}

inline void x86_memblt_long_aligned_dst(const void *src, void *dst, uint32 len)
{
	asm volatile (
   "leal	(%%edi,%%edi,2),%%ecx
	andl	$3,%%ecx
	subl	%%ecx,%%eax
	jle		__EndBytes
rep
	movsb
	movl	%%eax,%%ecx
	andl	$3,%%eax
	shrl	$2,%%ecx
rep
	movsl
__EndBytes:
	addl	%%eax,%%ecx
rep
	movsb"
	: "=S" (src), "=D" (dst), "=a" (len)
	: "0" (src), "1" (dst), "2" (len)
	: "cc", "%ecx" );
}
// above code by Paul Hsieh, parts by Ken Silverman

inline void mmx_3dn_memblt_long_long_aligned_dst(const void *src, void *dst, uint32 len)
{
	asm volatile (
   "leal	(%%edi,%%edi,2),%%ecx
	leal	(%%edi,%%ecx,2),%%ecx
	andl	$7,%%ecx
	subl	%%ecx,%%eax
	jle		__EndBytes
rep
	movsb
	movl	%%eax,%%ecx
	andl	$31,%%eax
	shrl	$5,%%ecx
	je		__EndBytes
	.align	16
__LongLong:
	prefetch	0x20(%%esi)		# a 3dn op
	movq	0x00(%%esi),%%mm0
	movq	0x08(%%esi),%%mm1
	movq	0x10(%%esi),%%mm2
	movq	0x18(%%esi),%%mm3
	movq	%%mm0,0x00(%%edi)
	movq	%%mm1,0x08(%%edi)
	movq	%%mm2,0x10(%%edi)
	movq	%%mm3,0x18(%%edi)
	addl	$32,%%esi
	addl	$32,%%edi
	dec		%%ecx
	jne		__LongLong
__EndBytes:
	addl	%%eax,%%ecx
rep
	movsb"
	: "=S" (src), "=D" (dst), "=a" (len)
	: "0" (src), "1" (dst), "2" (len)
	: "cc", "%st", "%st(1)", "%st(2)", "%st(3)", "%ecx" );
}
// warning: to avoid an emms, this function is to be used only at fpu TOS = 0,
// original code by Paul Hsieh and Ken Silverman, long_long adaptation by me

#endif // __INTEL__

// memblt()		: copies data chunks from the heap to the display buffer
//		- src,	const void *	: source location,		input
//		- dst,	void *			: destination,			input
//		- len,	uint32			: chunk size in bytes,	input
// returns
//		nil
// note
//		- PCI bus controllers and devices employ various transfer optimisation policies
//		  such as Write Combining (i.e. batching small-granularity, adjacent writes into
//		  whole-bus-words writes) and batching multi-word sequential accesses as bursts.
//		- consider host-originated-device-carried transfering where available (i.e. GART
//		  and similar scatter-geather techs), that would free the flipping thread from the
//		  blitting itself.
inline void memblt(const void *src, void *dst, uint32 len)
{
/*	::x86_memblt_long_aligned_src(src, dst, len);			*/
	// for WC-enabled platforms; this could alleviate the destination-alignment requirement
	// and allow us to focus on the source access alignment

	::x86_memblt_long_aligned_dst(src, dst, len);
	// for WC-disabled platforms; in this case the destination access alignment is determinant

/*	::mmx_3dn_memblt_long_long_aligned_dst(src, dst, len);	*/
	// for WC-disabled platforms capable of burst writes to the video

/*	::memcpy(dst, src, len);								*/
	// CRT's heap-to-heap could be appropriate for UMA platforms
}

#endif // __MEMBLT_H__
