#ifndef __RENDBASETYPES__
#define __RENDBASETYPES__

////////////////////////////////////////////////////////////////
// definitions
////////////////////////////////////////////////////////////////

typedef signed char		int8;
typedef unsigned char	uint8;

typedef short			int16;
typedef unsigned short	uint16;

typedef long			int32;
typedef unsigned long	uint32;

struct Rect
{
	float	left;
	float	top;
	float	right;
	float	bottom;
};

////////////////////////////////////////////////////////////////

struct Vertex2
{
	float x;
	float y;
};

struct Vertex4
{
	float x;
	float y;
	float z;
	float w;
};

struct Vertex4UV
{
	float x;
	float y;
	float z;
	float w;
	float u;
	float v;
};

struct Vertex4ARGB
{
	float x;
	float y;
	float z;
	float w;
	float a;
	float r;
	float g;
	float b;
};

#define IEEE754_SIGN_MASK					0x80000000UL

#define RENDER_STATE_MIPMAP_CAPACITY		16						// max num of levels of detail per mipmap
#define	RENDER_STATE_GRADIENT_VECTSIZE		6						// max num of elements per vertex sans the X, Y coordinates
#define	RENDER_STATE_SPANBUFF_VECTSIZE		2						// num of elements per vector in the span buffer

#define	RENDER_STATE_CULL_NONE				0
#define	RENDER_STATE_CULL_CW				~IEEE754_SIGN_MASK		// positive area - clockwise get culled
#define	RENDER_STATE_CULL_CCW				IEEE754_SIGN_MASK		// negative area - counter-clockwise get culled

#define	RENDER_STATE_VERT_INVALID			0
#define RENDER_STATE_VERT_BASIC				1						// X,Y,Z coordinates + RHW
#define	RENDER_STATE_VERT_UV				2						// X,Y,Z coordinates + RHW + U,V texture coordinates
#define	RENDER_STATE_VERT_ARGB				3						// X,Y,Z coordinates + RHW + A,R,G,B color coordinates

#define RENDER_STATE_OP_INVALID				0
#define	RENDER_STATE_OP_NOP					1						// dst.comp_i = dst.comp_i
#define	RENDER_STATE_OP_COPY				2						// dst.comp_i = src.comp_i
#define RENDER_STATE_OP_ADD					3						// dst.comp_1 = dst.comp_1 + src.comp_i
#define RENDER_STATE_OP_MODULATE			4						// dst.comp_i = dst.comp_i * src.comp_i
#define RENDER_STATE_OP_DOT3				5						// dst.comp_i = vect3(dst) * vect3(src)

#define RENDER_STATE_PIX_INVALID			0
#define	RENDER_STATE_PIX_ARGB1555			1						// A:1, R:5, G:5, B:5
#define RENDER_STATE_PIX_ARGB8888			2						// A:8, R:8, G:8, B:8
#define	RENDER_STATE_PIX_RGB565				3						//		R:5, G:6, B:5
#define RENDER_STATE_PIX_RGB888				4						//		R:8, G:8, B:8
#define RENDER_STATE_PIX_DA16				5						// Declination:16, Azimuth:16
#define RENDER_STATE_PIX_DA12				6						// Declination:12, Azimuth:12
#define RENDER_STATE_PIX_DA8				7						// Declination:8,  Azimuth:8

#define RENDER_STATE_SAMP_INVALID			0
#define RENDER_STATE_SAMP_CONST_NEAREST		1						// no mipmapping, nearest sampling
#define RENDER_STATE_SAMP_CONST_LINEAR		2						// no mipmapping, linear sampling
#define RENDER_STATE_SAMP_NEAREST_NEAREST	3						// nearest LOD, nearest sampling
#define RENDER_STATE_SAMP_NEAREST_LINEAR	4						// nearest LOD, linear sampling
#define RENDER_STATE_SAMP_LINEAR_NEAREST	5						// linear LOD, nearest sampling
#define RENDER_STATE_SAMP_LINEAR_LINEAR		6						// linear LOD, linear sampling

struct PixelBuffer
{
	int8 *					bits;
	int8 *					cursor;

	int32					dimx;
	int32					dimy;
	int32					row_stride;
	uint32					pix_format;		// one of RENDER_STATE_PIX_*
};

struct RenderState
{
	union
	{
		PixelBuffer			tex_lod[RENDER_STATE_MIPMAP_CAPACITY];
		PixelBuffer			tex;
	};
	int32					tex_lod_min;	// non-negative
	int32					tex_lod_max;	// non-negative
	float					tex_lod_bias;

	PixelBuffer				frame;
	PixelBuffer				aux;

	uint32					culling;		// one of RENDER_STATE_CULL_*
	uint32					vert_format;	// one of RENDER_STATE_VERT_*
	uint32					color_op;		// one of RENDER_STATE_OP_*
	uint32					sampling;		// one of RENDER_STATE_SAMP_*

	float					trigrad_x[RENDER_STATE_GRADIENT_VECTSIZE];
	float					trigrad_y[RENDER_STATE_GRADIENT_VECTSIZE];
};

struct Viewport
{
	Rect					screen;			// mapping onto screen; borders are inclusive
	float					depth_range_n;	// mapping onto depth range - mapping of the near clipping plane
	float					depth_range_f;	// mapping onto depth range - mapping of the far clipping plane
};

struct SpanBuffer
{
	float					(*column)[RENDER_STATE_SPANBUFF_VECTSIZE];	// num of columns must be greater than viewport width

	int32					x0;
	int32					x1;
};

////////////////////////////////////////////////////////////////
// declarations
////////////////////////////////////////////////////////////////

extern RenderState			rend_state;
extern Viewport				viewport;
extern SpanBuffer			span_above;

#endif // __RENDBASETYPES__
