#ifndef __RENDLIGHT__
#define __RENDLIGHT__

#include <math.h>

struct Color
{
	float					a;
	float					r;
	float					g;
	float					b;

	Color() : a(1.f), r(0.f), g(0.f), b(0.f)
	{ }

	Color(float arg_a,
		  float arg_r = 0.f,
		  float arg_g = 0.f,
		  float arg_b = 0.f)
	{
		a = arg_a;
		r = arg_r;
		g = arg_g;
		b = arg_b;
	}
};

struct Material
{
	Color					ambient;
	Color					diffuse;
	Color					specular;
	Color					emissive;
	float					shininess;			// range: [0, 128]
	
	Material() : ambient(1.f, .2f, .2f, .2f),
				 diffuse(1.f, .8f, .8f, .8f),
				 shininess(0.f)
	{ }
};

struct LightSource
{
	Color					ambient;
	Color					diffuse;
	Color					specular;
	float 					position[4];
	float					spot_direction[3];
	float					spot_exponent;		// range: [0, 128]
	float					spot_cutoff;		// range: [0, pi / 2], pi
	float					attenu_constant;	// range: [0, +inf)
	float					attenu_linear;		// range: [0, +inf)
	float					attenu_quadratic;	// range: [0, +inf)

	LightSource() : diffuse(1.f, 1.f, 1.f, 1.f),
					spot_exponent(0.f),
					spot_cutoff(M_PI),
					attenu_constant(1.f),
					attenu_linear(0.f),
					attenu_quadratic(0.f)
	{
		position[0] = 0.f;
		position[1] = 0.f;
		position[2] = 1.f;
		position[3] = 0.f;
		
		spot_direction[0] = 0.f;
		spot_direction[1] = 0.f;
		spot_direction[2] = -1.f;
	}
};

struct Illumination
{
	int						num_src;
	LightSource	*			src;
	Color					ambient;

	bool					local_viewer;
	bool					renorm;
	
	Illumination() : num_src(0),
					 src(NULL),
					 ambient(1.f, .2f, .2f, .2f),
					 local_viewer(false),
					 renorm(false)
	{ }
};

void LightVertex(const Illumination &, const Material &, const float (&)[4][4], const float (&)[4][4],
				 /*const*/ float (&)[4], /*const*/ float (&)[3], Color &);

void LightVertex(const Illumination &, const Material &, const float (&)[4][4], const float (&)[4][4],
				 /*const*/ float (&)[4], /*const*/ float (&)[3], Color &, Color &);

#endif // __RENDLIGHT__
