#ifndef __RENDPIXEL__
#define __RENDPIXEL__

#include <memory.h>
#include "rendBaseTypes.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// class Pixel
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////
// remark on code redundancy
//
// since some compilers (e.g. gcc 2.9-20000224) do not support zero-width bit-fields, we here define
// several base classes to accomodate for the several numbers of color channels
////////////////////////////////////////////////////////////////////////////////////////////////////

template < typename	INTEGRAL_T,
		   unsigned	NUMBITS_C3_T,
		   unsigned	NUMBITS_C2_T,
		   unsigned	NUMBITS_C1_T,
		   unsigned	NUMBITS_C0_T >
class Pixel4
{
public:

	INTEGRAL_T		c0 : NUMBITS_C0_T;
	INTEGRAL_T		c1 : NUMBITS_C1_T;
	INTEGRAL_T		c2 : NUMBITS_C2_T;
	INTEGRAL_T		c3 : NUMBITS_C3_T;

	static unsigned long	PackedSize()	{ return NUMBITS_C0_T + NUMBITS_C1_T + NUMBITS_C2_T + NUMBITS_C3_T + 7 >> 3; }

	void		SetPacked(const INTEGRAL_T packed)	{ ::memcpy(this, &packed, PackedSize()); }
	INTEGRAL_T	GetPacked()	const					{ return *(const INTEGRAL_T *)this; }

	void Set(const INTEGRAL_T c3,
			 const INTEGRAL_T c2,
			 const INTEGRAL_T c1,
			 const INTEGRAL_T c0)
	{
		SetPacked((c3 & (1 << NUMBITS_C3_T) - 1) << NUMBITS_C2_T + NUMBITS_C1_T + NUMBITS_C0_T |
				  (c2 & (1 << NUMBITS_C2_T) - 1) << NUMBITS_C1_T + NUMBITS_C0_T |
				  (c1 & (1 << NUMBITS_C1_T) - 1) << NUMBITS_C0_T |
				  (c0 & (1 << NUMBITS_C0_T) - 1));
	}

	Pixel4() {}
	Pixel4(const INTEGRAL_T c3,
		   const INTEGRAL_T c2,
		   const INTEGRAL_T c1,
		   const INTEGRAL_T c0)
	{
		Set(c3, c2, c1, c0);
	}
	
	Pixel4< INTEGRAL_T, NUMBITS_C3_T, NUMBITS_C2_T, NUMBITS_C1_T, NUMBITS_C0_T > &
	operator = (const Pixel4< INTEGRAL_T, NUMBITS_C3_T, NUMBITS_C2_T, NUMBITS_C1_T, NUMBITS_C0_T > &p)
	{
		SetPacked(p.GetPacked());
		return *this;
	}
};

template < typename	INTEGRAL_T,
		   unsigned	NUMBITS_C2_T,
		   unsigned	NUMBITS_C1_T,
		   unsigned	NUMBITS_C0_T >
class Pixel3
{
public:

	INTEGRAL_T		c0 : NUMBITS_C0_T;
	INTEGRAL_T		c1 : NUMBITS_C1_T;
	INTEGRAL_T		c2 : NUMBITS_C2_T;

	static unsigned long	PackedSize()	{ return NUMBITS_C0_T + NUMBITS_C1_T + NUMBITS_C2_T + 7 >> 3; }

	void		SetPacked(const INTEGRAL_T packed)	{ ::memcpy(this, &packed, PackedSize()); }
	INTEGRAL_T	GetPacked()	const					{ return *(const INTEGRAL_T *)this; }

	void Set(const INTEGRAL_T c2,
			 const INTEGRAL_T c1,
			 const INTEGRAL_T c0)
	{
		SetPacked((c2 & (1 << NUMBITS_C2_T) - 1) << NUMBITS_C1_T + NUMBITS_C0_T |
				  (c1 & (1 << NUMBITS_C1_T) - 1) << NUMBITS_C0_T |
				  (c0 & (1 << NUMBITS_C0_T) - 1));
	}

	Pixel3() {}
	Pixel3(const INTEGRAL_T c2,
		   const INTEGRAL_T c1,
		   const INTEGRAL_T c0)
	{
		Set(c2, c1, c0);
	}

	Pixel3< INTEGRAL_T, NUMBITS_C2_T, NUMBITS_C1_T, NUMBITS_C0_T > &
	operator = (const Pixel3< INTEGRAL_T, NUMBITS_C2_T, NUMBITS_C1_T, NUMBITS_C0_T > &p)
	{
		SetPacked(p.GetPacked());
		return *this;
	}
};

template < typename	INTEGRAL_T,
		   unsigned	NUMBITS_C1_T,
		   unsigned	NUMBITS_C0_T >
class Pixel2
{
public:

	INTEGRAL_T		c0 : NUMBITS_C0_T;
	INTEGRAL_T		c1 : NUMBITS_C1_T;

	static unsigned long	PackedSize()	{ return NUMBITS_C0_T + NUMBITS_C1_T + 7 >> 3; }

	void		SetPacked(const INTEGRAL_T packed)	{ ::memcpy(this, &packed, PackedSize()); }
	INTEGRAL_T	GetPacked()	const					{ return *(const INTEGRAL_T *)this; }

	void Set(const INTEGRAL_T c1,
			 const INTEGRAL_T c0)
	{
		SetPacked((c1 & (1 << NUMBITS_C1_T) - 1) << NUMBITS_C0_T |
				  (c0 & (1 << NUMBITS_C0_T) - 1));
	}

	Pixel2() {}
	Pixel2(const INTEGRAL_T c1,
		   const INTEGRAL_T c0)
	{
		Set(c1, c0);
	}

	Pixel2< INTEGRAL_T, NUMBITS_C1_T, NUMBITS_C0_T > &
	operator = (const Pixel2< INTEGRAL_T, NUMBITS_C1_T, NUMBITS_C0_T > &p)
	{
		SetPacked(p.GetPacked());
		return *this;
	}
};

////////////////////////////////////////////////////////////////////////////////////////////////////
// formatted pixel definitions
////////////////////////////////////////////////////////////////////////////////////////////////////

typedef Pixel4< uint32, 1,	5,	5,	5 >		PixelARGB1555;
typedef Pixel4< uint32, 8,	8,	8,	8 >		PixelARGB8888;

typedef Pixel3< uint32, 5,	6,	5 >			PixelRGB565;
typedef Pixel3< uint32, 8,	8,	8 >			PixelRGB888;

typedef Pixel2< uint32, 16, 16 >			PixelDA16;
typedef Pixel2< uint32, 12, 12 >			PixelDA12;
typedef Pixel2< uint32, 8,  8 >				PixelDA8;

////////////////////////////////////////////////////////////////////////////////////////////////////
// function definitions
////////////////////////////////////////////////////////////////////////////////////////////////////

template < typename	D_INTEGRAL_T,
		   unsigned	D_NUMBITS_A_T,
		   unsigned	D_NUMBITS_R_T,
		   unsigned	D_NUMBITS_G_T,
		   unsigned	D_NUMBITS_B_T,
		   typename	S_INTEGRAL_T,
		   unsigned	S_NUMBITS_A_T,
		   unsigned	S_NUMBITS_R_T,
		   unsigned	S_NUMBITS_G_T,
		   unsigned	S_NUMBITS_B_T >

inline Pixel4<D_INTEGRAL_T, D_NUMBITS_A_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>
		ConvertARGB(const Pixel4<S_INTEGRAL_T, S_NUMBITS_A_T, S_NUMBITS_R_T, S_NUMBITS_G_T, S_NUMBITS_B_T> src)
{
	const int32 a = ToIntRound((float) (int32) src.c3 * ((float) ((1 << D_NUMBITS_A_T) - 1) / (float) ((1 << S_NUMBITS_A_T) - 1)));
	const int32 r = ToIntRound((float) (int32) src.c2 * ((float) ((1 << D_NUMBITS_R_T) - 1) / (float) ((1 << S_NUMBITS_R_T) - 1)));
	const int32 g = ToIntRound((float) (int32) src.c1 * ((float) ((1 << D_NUMBITS_G_T) - 1) / (float) ((1 << S_NUMBITS_G_T) - 1)));
	const int32 b = ToIntRound((float) (int32) src.c0 * ((float) ((1 << D_NUMBITS_B_T) - 1) / (float) ((1 << S_NUMBITS_B_T) - 1)));

	return Pixel4< D_INTEGRAL_T, D_NUMBITS_A_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T >((D_INTEGRAL_T) a,
																							  (D_INTEGRAL_T) r,
																							  (D_INTEGRAL_T) g,
																							  (D_INTEGRAL_T) b);
}

template < typename	D_INTEGRAL_T,
		   unsigned	D_NUMBITS_R_T,
		   unsigned	D_NUMBITS_G_T,
		   unsigned	D_NUMBITS_B_T,
		   typename	S_INTEGRAL_T,
		   unsigned	S_NUMBITS_R_T,
		   unsigned	S_NUMBITS_G_T,
		   unsigned	S_NUMBITS_B_T >

inline Pixel3<D_INTEGRAL_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>
		ConvertRGB(const Pixel3<S_INTEGRAL_T, S_NUMBITS_R_T, S_NUMBITS_G_T, S_NUMBITS_B_T> src)
{
	const int32 r = ToIntRound((float) (int32) src.c2 * ((float) ((1 << D_NUMBITS_R_T) - 1) / (float) ((1 << S_NUMBITS_R_T) - 1)));
	const int32 g = ToIntRound((float) (int32) src.c1 * ((float) ((1 << D_NUMBITS_G_T) - 1) / (float) ((1 << S_NUMBITS_G_T) - 1)));
	const int32 b = ToIntRound((float) (int32) src.c0 * ((float) ((1 << D_NUMBITS_B_T) - 1) / (float) ((1 << S_NUMBITS_B_T) - 1)));

	return Pixel3<D_INTEGRAL_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>((D_INTEGRAL_T) r,
																			 (D_INTEGRAL_T) g,
																			 (D_INTEGRAL_T) b);
}

template < typename	D_INTEGRAL_T,
		   unsigned	D_NUMBITS_R_T,
		   unsigned	D_NUMBITS_G_T,
		   unsigned	D_NUMBITS_B_T,
		   typename	S_INTEGRAL_T,
		   unsigned	S_NUMBITS_A_T,
		   unsigned	S_NUMBITS_R_T,
		   unsigned	S_NUMBITS_G_T,
		   unsigned	S_NUMBITS_B_T >

inline Pixel3<D_INTEGRAL_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>
		ConvertARGB2RGB(const Pixel4<S_INTEGRAL_T, S_NUMBITS_A_T, S_NUMBITS_R_T, S_NUMBITS_G_T, S_NUMBITS_B_T> src)
{
	const int32 r = ToIntRound((float) (int32) src.c2 * ((float) ((1 << D_NUMBITS_R_T) - 1) / (float) ((1 << S_NUMBITS_R_T) - 1)));
	const int32 g = ToIntRound((float) (int32) src.c1 * ((float) ((1 << D_NUMBITS_G_T) - 1) / (float) ((1 << S_NUMBITS_G_T) - 1)));
	const int32 b = ToIntRound((float) (int32) src.c0 * ((float) ((1 << D_NUMBITS_B_T) - 1) / (float) ((1 << S_NUMBITS_B_T) - 1)));

	return Pixel3<D_INTEGRAL_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>((D_INTEGRAL_T) r,
																			 (D_INTEGRAL_T) g,
																			 (D_INTEGRAL_T) b);
}

template < typename	D_INTEGRAL_T,
		   unsigned	D_NUMBITS_A_T,
		   unsigned	D_NUMBITS_R_T,
		   unsigned	D_NUMBITS_G_T,
		   unsigned	D_NUMBITS_B_T,
		   typename	S_INTEGRAL_T,
		   unsigned	S_NUMBITS_R_T,
		   unsigned	S_NUMBITS_G_T,
		   unsigned	S_NUMBITS_B_T >

inline Pixel4<D_INTEGRAL_T, D_NUMBITS_A_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>
		ConvertRGB2ARGB(const Pixel3<S_INTEGRAL_T, S_NUMBITS_R_T, S_NUMBITS_G_T, S_NUMBITS_B_T> src)
{
	const int32 a = (1 << D_NUMBITS_A_T) - 1;
	const int32 r = ToIntRound((float) (int32) src.c2 * ((float) ((1 << D_NUMBITS_R_T) - 1) / (float) ((1 << S_NUMBITS_R_T) - 1)));
	const int32 g = ToIntRound((float) (int32) src.c1 * ((float) ((1 << D_NUMBITS_G_T) - 1) / (float) ((1 << S_NUMBITS_G_T) - 1)));
	const int32 b = ToIntRound((float) (int32) src.c0 * ((float) ((1 << D_NUMBITS_B_T) - 1) / (float) ((1 << S_NUMBITS_B_T) - 1)));

	return Pixel4<D_INTEGRAL_T, D_NUMBITS_A_T, D_NUMBITS_R_T, D_NUMBITS_G_T, D_NUMBITS_B_T>((D_INTEGRAL_T) a,
																							(D_INTEGRAL_T) r,
																							(D_INTEGRAL_T) g,
																							(D_INTEGRAL_T) b);
}

template < typename	D_INTEGRAL_T,
		   unsigned	D_NUMBITS_D_T,
		   unsigned	D_NUMBITS_A_T,
		   typename	S_INTEGRAL_T,
		   unsigned	S_NUMBITS_D_T,
		   unsigned	S_NUMBITS_A_T >

inline Pixel2<D_INTEGRAL_T, D_NUMBITS_D_T, D_NUMBITS_A_T>
		ConvertDA(const Pixel2<S_INTEGRAL_T, S_NUMBITS_D_T, S_NUMBITS_A_T> src)
{
	const int32 d = ToIntRound((float) (int32) src.c1 * ((float) ((1 << D_NUMBITS_D_T) - 1) / (float) ((1 << S_NUMBITS_D_T) - 1)));
	const int32 a = ToIntRound((float) (int32) src.c0 * ((float) ((1 << D_NUMBITS_A_T) - 1) / (float) ((1 << S_NUMBITS_A_T) - 1)));

	return Pixel2<D_INTEGRAL_T, D_NUMBITS_D_T, D_NUMBITS_A_T>((D_INTEGRAL_T) d,
															  (D_INTEGRAL_T) a);
}

#endif // __RENDPIXEL__
