#ifndef __RENDPRIM__
#define __RENDPRIM__

////////////////////////////////////////////////////////////////////////////////////////////////////
// prototypes
////////////////////////////////////////////////////////////////////////////////////////////////////

void PlotTriangle4(const float (&)[4][4], /*const*/ float (&)[4], /*const*/ float (&)[4], /*const*/ float (&)[4]);
void PlotTriangle4(/*const*/ float (&)[4], /*const*/ float (&)[4], /*const*/ float (&)[4]);

void PlotTriangle6(const float (&)[4][4], /*const*/ float (&)[6], /*const*/ float (&)[6], /*const*/ float (&)[6]);
void PlotTriangle6(/*const*/ float (&)[6], /*const*/ float (&)[6], /*const*/ float (&)[6]);

void PlotTriangle8(const float (&)[4][4], /*const*/ float (&)[8], /*const*/ float (&)[8], /*const*/ float (&)[8]);
void PlotTriangle8(/*const*/ float (&)[8], /*const*/ float (&)[8], /*const*/ float (&)[8]);

////////////////////////////////////////////////////////////////////////////////////////////////////
// inlined definitions
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////
// remark on conventions
//
// perspective division of the positional coordinates from clip space (past our projection
// transformation) practically stretches the view volume into a cube, one centered over the
// origin and spanning a unit in each direction of the axes. coords within said cubic volume
// are known as normalized device coordinates (NDC)
////////////////////////////////////////////////////////////////////////////////////////////////////

inline void NDCToScreen(float (&v)[3])
{
	// map the normalized device x and y coordinates to screen coordinates
	v[0] = v[0] * (viewport.screen.right - viewport.screen.left) * .5f +
				  (viewport.screen.right + viewport.screen.left) * .5f;

	v[1] = v[1] * (viewport.screen.top - viewport.screen.bottom) * .5f +
				  (viewport.screen.top + viewport.screen.bottom) * .5f;

	// map the normalized device z coordinate to the specified depth range
	v[2] = v[2] * (viewport.depth_range_f - viewport.depth_range_n) * .5f +
				  (viewport.depth_range_f + viewport.depth_range_n) * .5f;
}

// MapToScreen()	: maps a vector from clipping space to screen space
//		VECTSIZE_T,	int							: must be of R4 or greater
//		- vi,		const float (&)[VECTSIZE_T]	: vector in clipping space,	input
//		- vo,		float (&) [VECTSIZE_T]		: vector in screen space,	output
// returns
//		nil
// note
//		- output vector's 4th component is the reciprocal homogeneous W
template < int VECTSIZE_T >
inline void MapToScreen(const float (&vi)[VECTSIZE_T], float (&vo)[VECTSIZE_T])
{
	// perform perspective division of the clip-space vector
	const float oow = 1.f / vi[3];

	for (int comp = 0; comp < VECTSIZE_T; comp++)
		vo[comp] = vi[comp] * oow;

	vo[3] = oow;

	NDCToScreen(*(float (*)[3])&vo);
}

#endif // __RENDPRIM__
