#include <math.h>
#include <mathinline.h>
#include "rendSpan.h"
#include "rendVect.h"
#include "rendPixel.h"
#include "rendHelper.h"

inline void __unused_float(const float) {}

// Span4_nop()	: scan-converts a horizontal span into the void
//		- const float (&)[4]	: span 1st vertex,	input
//		- const float (&)[4]	: span 2nd vertex,	input
// returns
//		nil
void Span4_nop(const float (&)[4],
			   const float (&)[4])
{ }

// Span6_nop()	: scan-converts a horizontal span into the void
//		- const float (&)[6]	: span 1st vertex,	input
//		- const float (&)[6]	: span 2nd vertex,	input
// returns
//		nil
void Span6_nop(const float (&)[6],
			   const float (&)[6])
{ }

// Span8_nop()	: scan-converts a horizontal span into the void
//		- const float (&)[8]	: span 1st vertex,	input
//		- const float (&)[8]	: span 2nd vertex,	input
// returns
//		nil
void Span8_nop(const float (&)[8],
			   const float (&)[8])
{ }


template < int TARGET_PIX_FORMAT_T >
inline void TexCpyNearest(const int int_x0,
						  const int i,
						  const float ui,
						  const float vi,
						  const int lod)
{
	int snap_ui = ToIntRound(ui);
	int snap_vi = ToIntRound(vi);

	snap_ui %= rend_state.tex_lod[lod].dimx;
	snap_vi %= rend_state.tex_lod[lod].dimy;

	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	case RENDER_STATE_PIX_ARGB8888:
	case RENDER_STATE_PIX_RGB565:
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelARGB8888 tex = *(const PixelARGB8888 *)(rend_state.tex_lod[lod].bits +
																snap_vi * rend_state.tex_lod[lod].row_stride +
																snap_ui * PixelARGB8888::PackedSize());
		switch (TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_ARGB1555:

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = ConvertARGB< uint32, 1, 5, 5, 5 >(tex);
			break;

		case RENDER_STATE_PIX_ARGB8888:

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = tex;
			break;

		case RENDER_STATE_PIX_RGB565:

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = ConvertARGB2RGB< uint32, 5, 6, 5 >(tex);
			break;

		case RENDER_STATE_PIX_RGB888:

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = ConvertARGB2RGB< uint32, 8, 8, 8 >(tex);
			break;
		}
		break;
	}
	case RENDER_STATE_PIX_DA16:
	case RENDER_STATE_PIX_DA12:
	case RENDER_STATE_PIX_DA8:
	{
		const PixelDA16 tex = *(const PixelDA16 *)(rend_state.tex_lod[lod].bits +
														snap_vi * rend_state.tex_lod[lod].row_stride +
														snap_ui * PixelDA16::PackedSize());
		switch (TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_DA16:

			*(PixelDA16 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA16::PackedSize()) = tex;
			break;

		case RENDER_STATE_PIX_DA12:

			*(PixelDA12 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA12::PackedSize()) = ConvertDA< uint32, 12, 12 >(tex);
			break;

		case RENDER_STATE_PIX_DA8:

			*(PixelDA8 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA8::PackedSize()) = ConvertDA< uint32, 8, 8 >(tex);
			break;
		}
		break;
	} }
}

template < int TARGET_PIX_FORMAT_T >
inline void TexCpyLinear(const int int_x0,
						 const int i,
						 const float ui,
						 const float vi,
						 const int lod)
{
	const float snap_ui = floorf(ui);
	const float snap_vi = floorf(vi);

	const float mass_u1 = ui - snap_ui;
	const float mass_v1 = vi - snap_vi;

	const float mass00 = (1.f - mass_u1) * (1.f - mass_v1);
	const float mass01 = mass_u1		 * (1.f - mass_v1);
	const float mass10 = (1.f - mass_u1) * mass_v1;
	const float mass11 = mass_u1		 * mass_v1;

	int ui0 = ToInt(snap_ui);
	int vi0 = ToInt(snap_vi);
	int ui1 = ui0 + 1;
	int vi1 = vi0 + 1;

	ui0 %= rend_state.tex_lod[lod].dimx;
	vi0 %= rend_state.tex_lod[lod].dimy;
	ui1 %= rend_state.tex_lod[lod].dimx;
	vi1 %= rend_state.tex_lod[lod].dimy;

	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	case RENDER_STATE_PIX_ARGB8888:
	case RENDER_STATE_PIX_RGB565:
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelARGB8888 tex00 = *(const PixelARGB8888 *)(rend_state.tex_lod[lod].bits +
																vi0 * rend_state.tex_lod[lod].row_stride +
																ui0 * PixelARGB8888::PackedSize());
		const PixelARGB8888 tex01 = *(const PixelARGB8888 *)(rend_state.tex_lod[lod].bits +
																vi0 * rend_state.tex_lod[lod].row_stride +
																ui1 * PixelARGB8888::PackedSize());
		const PixelARGB8888 tex10 = *(const PixelARGB8888 *)(rend_state.tex_lod[lod].bits +
																vi1 * rend_state.tex_lod[lod].row_stride +
																ui0 * PixelARGB8888::PackedSize());
		const PixelARGB8888 tex11 = *(const PixelARGB8888 *)(rend_state.tex_lod[lod].bits +
																vi1 * rend_state.tex_lod[lod].row_stride +
																ui1 * PixelARGB8888::PackedSize());
		switch (TARGET_PIX_FORMAT_T) {

		case RENDER_STATE_PIX_ARGB1555:
		{
			const int32 a = ToIntRound(((float) tex00.c3 * mass00 + (float) tex01.c3 * mass01 + (float) tex10.c3 * mass10 + (float) tex11.c3 * mass11) * ( 1.f / 255.f));
			const int32 r = ToIntRound(((float) tex00.c2 * mass00 + (float) tex01.c2 * mass01 + (float) tex10.c2 * mass10 + (float) tex11.c2 * mass11) * (31.f / 255.f));
			const int32 g = ToIntRound(((float) tex00.c1 * mass00 + (float) tex01.c1 * mass01 + (float) tex10.c1 * mass10 + (float) tex11.c1 * mass11) * (31.f / 255.f));
			const int32 b = ToIntRound(((float) tex00.c0 * mass00 + (float) tex01.c0 * mass01 + (float) tex10.c0 * mass10 + (float) tex11.c0 * mass11) * (31.f / 255.f));

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = PixelARGB1555(a, r, g, b);
			break;
		}
		case RENDER_STATE_PIX_ARGB8888:
		{
			const int32 a = ToIntRound((float) tex00.c3 * mass00 + (float) tex01.c3 * mass01 + (float) tex10.c3 * mass10 + (float) tex11.c3 * mass11);
			const int32 r = ToIntRound((float) tex00.c2 * mass00 + (float) tex01.c2 * mass01 + (float) tex10.c2 * mass10 + (float) tex11.c2 * mass11);
			const int32 g = ToIntRound((float) tex00.c1 * mass00 + (float) tex01.c1 * mass01 + (float) tex10.c1 * mass10 + (float) tex11.c1 * mass11);
			const int32 b = ToIntRound((float) tex00.c0 * mass00 + (float) tex01.c0 * mass01 + (float) tex10.c0 * mass10 + (float) tex11.c0 * mass11);

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = PixelARGB8888(a, r, g, b);
			break;
		}
		case RENDER_STATE_PIX_RGB565:
		{
			const int32 r = ToIntRound(((float) tex00.c2 * mass00 + (float) tex01.c2 * mass01 + (float) tex10.c2 * mass10 + (float) tex11.c2 * mass11) * (31.f / 255.f));
			const int32 g = ToIntRound(((float) tex00.c1 * mass00 + (float) tex01.c1 * mass01 + (float) tex10.c1 * mass10 + (float) tex11.c1 * mass11) * (63.f / 255.f));
			const int32 b = ToIntRound(((float) tex00.c0 * mass00 + (float) tex01.c0 * mass01 + (float) tex10.c0 * mass10 + (float) tex11.c0 * mass11) * (31.f / 255.f));

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = PixelRGB565(r, g, b);
			break;
		}
		case RENDER_STATE_PIX_RGB888:
		{
			const int32 r = ToIntRound((float) tex00.c2 * mass00 + (float) tex01.c2 * mass01 + (float) tex10.c2 * mass10 + (float) tex11.c2 * mass11);
			const int32 g = ToIntRound((float) tex00.c1 * mass00 + (float) tex01.c1 * mass01 + (float) tex10.c1 * mass10 + (float) tex11.c1 * mass11);
			const int32 b = ToIntRound((float) tex00.c0 * mass00 + (float) tex01.c0 * mass01 + (float) tex10.c0 * mass10 + (float) tex11.c0 * mass11);

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = PixelRGB888(r, g, b);
			break;
		} }
		break;
	}
	case RENDER_STATE_PIX_DA16:
	case RENDER_STATE_PIX_DA12:
	case RENDER_STATE_PIX_DA8:
	{
		const PixelDA16 tex00 = *(const PixelDA16 *)(rend_state.tex_lod[lod].bits +
														vi0 * rend_state.tex_lod[lod].row_stride +
														ui0 * PixelDA16::PackedSize());
		const PixelDA16 tex01 = *(const PixelDA16 *)(rend_state.tex_lod[lod].bits +
														vi0 * rend_state.tex_lod[lod].row_stride +
														ui1 * PixelDA16::PackedSize());
		const PixelDA16 tex10 = *(const PixelDA16 *)(rend_state.tex_lod[lod].bits +
														vi1 * rend_state.tex_lod[lod].row_stride +
														ui0 * PixelDA16::PackedSize());
		const PixelDA16 tex11 = *(const PixelDA16 *)(rend_state.tex_lod[lod].bits +
														vi1 * rend_state.tex_lod[lod].row_stride +
														ui1 * PixelDA16::PackedSize());

		const float de00 = (float) (int32) tex00.c1 * (M_PI / 65535.f);
		const float az00 = (float) (int32) tex00.c0 * (2.f * M_PI / 65536.f);

		const float de01 = (float) (int32) tex01.c1 * (M_PI / 65535.f);
		const float az01 = (float) (int32) tex01.c0 * (2.f * M_PI / 65536.f);

		const float de10 = (float) (int32) tex10.c1 * (M_PI / 65535.f);
		const float az10 = (float) (int32) tex10.c0 * (2.f * M_PI / 65536.f);

		const float de11 = (float) (int32) tex11.c1 * (M_PI / 65535.f);
		const float az11 = (float) (int32) tex11.c0 * (2.f * M_PI / 65536.f);

		const float n00[3] = { sinf(de00) * cosf(az00),  sinf(de00) * sinf(az00),  cosf(de00) };
		const float n01[3] = { sinf(de01) * cosf(az01),  sinf(de01) * sinf(az01),  cosf(de01) };
		const float n10[3] = { sinf(de10) * cosf(az10),  sinf(de10) * sinf(az10),  cosf(de10) };
		const float n11[3] = { sinf(de11) * cosf(az11),  sinf(de11) * sinf(az11),  cosf(de11) };

		const vect<3> n = ~(mass00 * vect<3>::adopt(n00) +
							mass01 * vect<3>::adopt(n01) +
							mass10 * vect<3>::adopt(n10) +
							mass11 * vect<3>::adopt(n11));

		const float de = acosf(n[2]);
		const float az = atan2f(n[1] / sinf(de),
								n[0] / sinf(de));

		switch (TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_DA16:

			((PixelDA16 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA16::PackedSize()))->Set(ToIntRound(de * (65535.f / M_PI)),
																								   ToIntRound(az * (65536.f / (2.f * M_PI))));
			break;

		case RENDER_STATE_PIX_DA12:

			((PixelDA12 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA12::PackedSize()))->Set(ToIntRound(de * (4095.f / M_PI)),
																								   ToIntRound(az * (4096.f / (2.f * M_PI))));
			break;

		case RENDER_STATE_PIX_DA8:

			((PixelDA8 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA8::PackedSize()))->Set(ToIntRound(de * (255.f / M_PI)),
																								 ToIntRound(az * (256.f / (2.f * M_PI))));
			break;
		}
		break;
	} }
}

// Span4UV_cpy()				: scan-converts a texture-mapped horizontal span, op COPY, no mipmapping
//		TARGET_PIX_FORMAT_T			: one of RENDER_STATE_PIX_*
//		TEXTURE_SAMP_TYPE_T			: one of RENDER_STATE_SAMP_CONST_*
//		- p0,	const Vertex4UV &	: span 1st vertex,	input
//		- p1,	const Vertex4UV &	: span 2nd vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order

template < int TARGET_PIX_FORMAT_T,
		   int TEXTURE_SAMP_TYPE_T >

inline void Span4UV_cpy(const Vertex4UV &p0,
						const Vertex4UV &p1)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0	= snap_x0 - p0.x;

    const float dz_dx	= rend_state.trigrad_x[0];
    const float doow_dx = rend_state.trigrad_x[1];
    const float duow_dx = rend_state.trigrad_x[2];
    const float dvow_dx = rend_state.trigrad_x[3];

    const int span_len	= ToInt(snap_x1 - snap_x0) + 1;
    const int int_x0	= ToInt(snap_x0);

    for (int i = 0; i < span_len; i++)
    {
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx;	__unused_float(zi);
        const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
        const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
        const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;

        const float ui = uowi * wi;
        const float vi = vowi * wi;

		switch(TEXTURE_SAMP_TYPE_T)
		{
		case RENDER_STATE_SAMP_CONST_NEAREST:

			TexCpyNearest< TARGET_PIX_FORMAT_T >(int_x0, i, ui, vi, 0);
			break;

		case RENDER_STATE_SAMP_CONST_LINEAR:

			TexCpyLinear< TARGET_PIX_FORMAT_T >(int_x0, i, ui, vi, 0);
			break;
		}
    }
}
	
void Span4UV_cpy_ARGB8888_nearest(const Vertex4UV &p0,
								  const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_ARGB1555_nearest(const Vertex4UV &p0,
								  const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_RGB888_nearest(const Vertex4UV &p0,
								const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_RGB565_nearest(const Vertex4UV &p0,
								const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_ARGB8888_linear(const Vertex4UV &p0,
								 const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_ARGB1555_linear(const Vertex4UV &p0,
								 const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_RGB888_linear(const Vertex4UV &p0,
							   const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_RGB565_linear(const Vertex4UV &p0,
							   const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA16_nearest(const Vertex4UV &p0,
							  const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA16, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA12_nearest(const Vertex4UV &p0,
							  const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA12, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA8_nearest(const Vertex4UV &p0,
							 const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA8, RENDER_STATE_SAMP_CONST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA16_linear(const Vertex4UV &p0,
							 const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA16, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA12_linear(const Vertex4UV &p0,
							 const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA12, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA8_linear(const Vertex4UV &p0,
							const Vertex4UV &p1)
{
	Span4UV_cpy< RENDER_STATE_PIX_DA8, RENDER_STATE_SAMP_CONST_LINEAR >(p0, p1);
}


template < int TARGET_PIX_FORMAT_T,
		   int TEXTURE_SAMP_TYPE_T,
		   int RUN_T >
inline void TexCpyMipRun(const int int_x0,
						 int &i,
						 const int (&run_len)[3],
						 const Vertex4UV &p0,
						 const float err_x0,
						 const float dz_dx,
						 const float doow_dx,
						 const float duow_dx,
						 const float dvow_dx,
						 const float doow_dy,
						 const float duow_dy,
						 const float dvow_dy)
{
	for (; i < run_len[RUN_T]; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx; __unused_float(zi);
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float uowi = p0.u + (err_x0 + (float) i) * duow_dx;
		const float vowi = p0.v + (err_x0 + (float) i) * dvow_dx;

		const float wi = 1.f / oowi;
		float ui = uowi * wi;
		float vi = vowi * wi;

		if (1 != RUN_T)
		{
			// calculate the tex-mapping of a fictitious pixel above this one
			span_above.column[int_x0 + i + 1][0] = (uowi - duow_dy) / (oowi - doow_dy);
			span_above.column[int_x0 + i + 1][1] = (vowi - dvow_dy) / (oowi - doow_dy);
		}

		// calculate the derivatives dudx, dudy, dvdx and dvdy
		// as differences with the pixels on the left and above
		const float dudx = ui - span_above.column[int_x0 + i][0];
		const float dvdx = vi - span_above.column[int_x0 + i][1];
		const float dudy = ui - span_above.column[int_x0 + i + 1][0];
		const float dvdy = vi - span_above.column[int_x0 + i + 1][1];

		span_above.column[int_x0 + i + 1][0] = ui;
		span_above.column[int_x0 + i + 1][1] = vi;

		// calculate the texture scale factor (>1 is minification)
		const float tdx = dudx * dudx + dvdx * dvdx;
		const float tdy = dudy * dudy + dvdy * dvdy;
		const float rho = sqrtf(Max(tdx, tdy));

		// pick a texture LOD based on the scale factor
		int lod = ToIntRound(log2(rho) + rend_state.tex_lod_bias);

//		const float oolog2 = 1.f / logf(2);
//		int lod = ToIntRound(logf(rho) * oolog2 + rend_state.tex_lod_bias);

		if (lod < rend_state.tex_lod_min)
			lod = rend_state.tex_lod_min;
		else
		if (lod > rend_state.tex_lod_max)
			lod = rend_state.tex_lod_max;

		const float snap_rho = (float)(1 << lod);
		ui = (ui - .5f * (snap_rho - 1.f)) / snap_rho;
		vi = (vi - .5f * (snap_rho - 1.f)) / snap_rho;

		switch(TEXTURE_SAMP_TYPE_T)
		{
		case RENDER_STATE_SAMP_NEAREST_NEAREST:

			TexCpyNearest< TARGET_PIX_FORMAT_T >(int_x0, i, ui, vi, lod);
			break;

		case RENDER_STATE_SAMP_NEAREST_LINEAR:

			TexCpyLinear< TARGET_PIX_FORMAT_T >(int_x0, i, ui, vi, lod);
			break;
		}
	}
}

// Span4UV_cpy_mip_nearest()		: scan-converts a texture-mapped horizontal span, op COPY, nearest mipmap selection
//		TARGET_PIX_FORMAT_T			: one of RENDER_STATE_PIX_(A)RGB*
//		TEXTURE_SAMP_TYPE_T			: one of RENDER_STATE_SAMP_NEAREST_*
//		- p0,	const Vertex4UV &	: span 1st vertex,	input
//		- p1,	const Vertex4UV &	: span 2nd vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order

template < int TARGET_PIX_FORMAT_T,
		   int TEXTURE_SAMP_TYPE_T >

inline void Span4UV_cpy_mip_nearest(const Vertex4UV &p0,
									const Vertex4UV &p1)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0	= snap_x0 - p0.x;

	const float dz_dx	= rend_state.trigrad_x[0];
	const float doow_dx = rend_state.trigrad_x[1];
	const float duow_dx = rend_state.trigrad_x[2];
	const float dvow_dx = rend_state.trigrad_x[3];

	const float doow_dy = rend_state.trigrad_y[1];
	const float duow_dy = rend_state.trigrad_y[2];
	const float dvow_dy = rend_state.trigrad_y[3];

	const int32 span_len = ToInt(snap_x1 - snap_x0) + 1;
	const int32 int_x0	 = ToInt(snap_x0);
	const int32 int_x1	 = ToInt(snap_x1);

	// calculate the tex-mapping of a fictitious pixel on the left of the span
	// and store said mapping in the span-above buffer at position x0
	const float oow = p0.w + (err_x0 - 1.f) * doow_dx;
	const float uow = p0.u + (err_x0 - 1.f) * duow_dx;
	const float vow = p0.v + (err_x0 - 1.f) * dvow_dx;

	const float w = 1.f / oow;
	span_above.column[int_x0][0] = uow * w;
	span_above.column[int_x0][1] = vow * w;

	// determine what portion of the current span is covered by the previous span from the scanline above;
	// in the worst case this could produce three runs based on coverage - two uncovered ends and a covered middle
	const int run_len[3] =
	{
		Min(span_above.x0 - int_x0, span_len),	// length of the run that lies on the left of the span above
		Min(span_above.x1 - int_x0, span_len),	// combined length of the run on the left and the run below the span above
		int_x1 < span_above.x1 ? 0 : span_len	// combined length of all three runs (on the left, below and on the right)
	};

	span_above.x0 = int_x0;
	span_above.x1 = int_x1 + 1;	// keep the interval open from the right

	// process the three coverage runs sequentially, startig with the left uncovered run
	int i = 0;

	TexCpyMipRun< TARGET_PIX_FORMAT_T, TEXTURE_SAMP_TYPE_T, 0 >(int_x0, i, run_len,
																p0, err_x0, dz_dx,
																doow_dx, duow_dx, dvow_dx,
																doow_dy, duow_dy, dvow_dy);
	// process the middle covered run
	TexCpyMipRun< TARGET_PIX_FORMAT_T, TEXTURE_SAMP_TYPE_T, 1 >(int_x0, i, run_len,
																p0, err_x0, dz_dx,
																doow_dx, duow_dx, dvow_dx,
																doow_dy, duow_dy, dvow_dy);
	// process the right uncovered run
	TexCpyMipRun< TARGET_PIX_FORMAT_T, TEXTURE_SAMP_TYPE_T, 2 >(int_x0, i, run_len,
																p0, err_x0, dz_dx,
																doow_dx, duow_dx, dvow_dx,
																doow_dy, duow_dy, dvow_dy);
}

void Span4UV_cpy_ARGB8888_nearest_nearest(const Vertex4UV &p0,
										  const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_ARGB1555_nearest_nearest(const Vertex4UV &p0,
										  const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_RGB888_nearest_nearest(const Vertex4UV &p0,
										const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_RGB888, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_RGB565_nearest_nearest(const Vertex4UV &p0,
										const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_RGB565, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_ARGB8888_nearest_linear(const Vertex4UV &p0,
										 const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_ARGB1555_nearest_linear(const Vertex4UV &p0,
										 const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_RGB888_nearest_linear(const Vertex4UV &p0,
									   const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_RGB888, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_RGB565_nearest_linear(const Vertex4UV &p0,
									   const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_RGB565, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA16_nearest_nearest(const Vertex4UV &p0,
									  const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA16, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA12_nearest_nearest(const Vertex4UV &p0,
									  const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA12, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA8_nearest_nearest(const Vertex4UV &p0,
									 const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA8, RENDER_STATE_SAMP_NEAREST_NEAREST >(p0, p1);
}

void Span4UV_cpy_DA16_nearest_linear(const Vertex4UV &p0,
									 const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA16, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA12_nearest_linear(const Vertex4UV &p0,
									 const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA12, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}

void Span4UV_cpy_DA8_nearest_linear(const Vertex4UV &p0,
									const Vertex4UV &p1)
{
	Span4UV_cpy_mip_nearest< RENDER_STATE_PIX_DA8, RENDER_STATE_SAMP_NEAREST_LINEAR >(p0, p1);
}


template < int TARGET_PIX_FORMAT_T >
inline void ColorCpy(const int int_x0,
					 const int i,
					 const float ai,
					 const float ri,
					 const float gi,
					 const float bi)
{
	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	{
		const int32 a = ToIntRound( 1.f * ai);
		const int32 r = ToIntRound(31.f * ri);
		const int32 g = ToIntRound(31.f * gi);
		const int32 b = ToIntRound(31.f * bi);

		*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = PixelARGB1555(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_ARGB8888:
	{
		const int32 a = ToIntRound(255.f * ai);
		const int32 r = ToIntRound(255.f * ri);
		const int32 g = ToIntRound(255.f * gi);
		const int32 b = ToIntRound(255.f * bi);

		*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = PixelARGB8888(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB565:
	{
		const int32 r = ToIntRound(31.f * ri);
		const int32 g = ToIntRound(63.f * gi);
		const int32 b = ToIntRound(31.f * bi);

		*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = PixelRGB565(r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB888:
	{
		const int32 r = ToIntRound(255.f * ri);
		const int32 g = ToIntRound(255.f * gi);
		const int32 b = ToIntRound(255.f * bi);

		*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = PixelRGB888(r, g, b);
		break;
	} }
}

template < int TARGET_PIX_FORMAT_T >
inline void ColorAdd(const int int_x0,
					 const int i,
					 const float ai,
					 const float ri,
					 const float gi,
					 const float bi)
{
	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	{
		const PixelARGB1555 addend = *(const PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize());

		const int32 a = Min(Max(ToIntRound((float) addend.c3 +  1.f * ai), 0L),  1L);
		const int32 r = Min(Max(ToIntRound((float) addend.c2 + 31.f * ri), 0L), 31L);
		const int32 g = Min(Max(ToIntRound((float) addend.c1 + 31.f * gi), 0L), 31L);
		const int32 b = Min(Max(ToIntRound((float) addend.c0 + 31.f * bi), 0L), 31L);

		*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = PixelARGB1555(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_ARGB8888:
	{
		const PixelARGB8888 addend = *(const PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize());

		const int32 a = Min(Max(ToIntRound((float) addend.c3 + 255.f * ai), 0L), 255L);
		const int32 r = Min(Max(ToIntRound((float) addend.c2 + 255.f * ri), 0L), 255L);
		const int32 g = Min(Max(ToIntRound((float) addend.c1 + 255.f * gi), 0L), 255L);
		const int32 b = Min(Max(ToIntRound((float) addend.c0 + 255.f * bi), 0L), 255L);

		*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = PixelARGB8888(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB565:
	{
		const PixelRGB565 addend = *(const PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize());

		const int32 r = Min(Max(ToIntRound((float) addend.c2 + 31.f * ri), 0L), 31L);
		const int32 g = Min(Max(ToIntRound((float) addend.c1 + 63.f * gi), 0L), 63L);
		const int32 b = Min(Max(ToIntRound((float) addend.c0 + 31.f * bi), 0L), 31L);

		*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = PixelRGB565(r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelRGB888 addend = *(const PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize());

		const int32 r = Min(Max(ToIntRound((float) addend.c2 + 255.f * ri), 0L), 255L);
		const int32 g = Min(Max(ToIntRound((float) addend.c1 + 255.f * gi), 0L), 255L);
		const int32 b = Min(Max(ToIntRound((float) addend.c0 + 255.f * bi), 0L), 255L);

		*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = PixelRGB888(r, g, b);
		break;
	} }
}

template < int TARGET_PIX_FORMAT_T >
inline void ColorMod(const int int_x0,
					 const int i,
					 const float ai,
					 const float ri,
					 const float gi,
					 const float bi)
{
	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	{
		const PixelARGB1555 multiplicand = *(const PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize());

		const int32 a = ToIntRound((float) multiplicand.c3 * ai);
		const int32 r = ToIntRound((float) multiplicand.c2 * ri);
		const int32 g = ToIntRound((float) multiplicand.c1 * gi);
		const int32 b = ToIntRound((float) multiplicand.c0 * bi);

		*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = PixelARGB1555(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_ARGB8888:
	{
		const PixelARGB8888 multiplicand = *(const PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize());

		const int32 a = ToIntRound((float) multiplicand.c3 * ai);
		const int32 r = ToIntRound((float) multiplicand.c2 * ri);
		const int32 g = ToIntRound((float) multiplicand.c1 * gi);
		const int32 b = ToIntRound((float) multiplicand.c0 * bi);

		*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = PixelARGB8888(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB565:
	{
		const PixelRGB565 multiplicand = *(const PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize());

		const int32 r = ToIntRound((float) multiplicand.c2 * ri);
		const int32 g = ToIntRound((float) multiplicand.c1 * gi);
		const int32 b = ToIntRound((float) multiplicand.c0 * bi);

		*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = PixelRGB565(r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelRGB888 multiplicand = *(const PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize());

		const int32 r = ToIntRound((float) multiplicand.c2 * ri);
		const int32 g = ToIntRound((float) multiplicand.c1 * gi);
		const int32 b = ToIntRound((float) multiplicand.c0 * bi);

		*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = PixelRGB888(r, g, b);
		break;
	} }
}

template < int TARGET_PIX_FORMAT_T >
inline void ColorDot(const int int_x0,
					 const int i,
					 const float ai,
					 const float ri,
					 const float gi,
					 const float bi)
{
	switch(TARGET_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	{
		const PixelDA8 src = *(PixelDA8 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA8::PackedSize());

		const float de = (float) (int32) src.c1 * (M_PI / 255.f);
		const float az = (float) (int32) src.c0 * (2.f * M_PI / 256.f);

		const float sx = sinf(de) * cosf(az);
		const float sy = sinf(de) * sinf(az);
		const float sz = cosf(de);

		const float dot = Max(sx * ri + sy * gi + sz * bi, 0.f);

		const int32 a = ToIntRound( 1.f * dot);
		const int32 r = ToIntRound(31.f * dot);
		const int32 g = ToIntRound(31.f * dot);
		const int32 b = ToIntRound(31.f * dot);

		*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = PixelARGB1555(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_ARGB8888:
	{
		const PixelDA16 src = *(PixelDA16 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA16::PackedSize());

		const float de = (float) (int32) src.c1 * (M_PI / 65535.f);
		const float az = (float) (int32) src.c0 * (2.f * M_PI / 65536.f);

		const float sx = sinf(de) * cosf(az);
		const float sy = sinf(de) * sinf(az);
		const float sz = cosf(de);

		const float dot = Max(sx * ri + sy * gi + sz * bi, 0.f);

		const int32 a = ToIntRound(255.f * dot);
		const int32 r = ToIntRound(255.f * dot);
		const int32 g = ToIntRound(255.f * dot);
		const int32 b = ToIntRound(255.f * dot);

		*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = PixelARGB8888(a, r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB565:
	{
		const PixelDA8 src = *(PixelDA8 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA8::PackedSize());

		const float de = (float) (int32) src.c1 * (M_PI / 255.f);
		const float az = (float) (int32) src.c0 * (2.f * M_PI / 256.f);

		const float sx = sinf(de) * cosf(az);
		const float sy = sinf(de) * sinf(az);
		const float sz = cosf(de);

		const float dot = Max(sx * ri + sy * gi + sz * bi, 0.f);

		const int32 r = ToIntRound(31.f * dot);
		const int32 g = ToIntRound(63.f * dot);
		const int32 b = ToIntRound(31.f * dot);

		*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = PixelRGB565(r, g, b);
		break;
	}
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelDA12 src = *(PixelDA12 *)(rend_state.frame.cursor + (int_x0 + i) * PixelDA12::PackedSize());

		const float de = (float) (int32) src.c1 * (M_PI / 4095.f);
		const float az = (float) (int32) src.c0 * (2.f * M_PI / 4096.f);

		const float sx = sinf(de) * cosf(az);
		const float sy = sinf(de) * sinf(az);
		const float sz = cosf(de);

		const float dot = Max(sx * ri + sy * gi + sz * bi, 0.f);

		const int32 r = ToIntRound(255.f * dot);
		const int32 g = ToIntRound(255.f * dot);
		const int32 b = ToIntRound(255.f * dot);

		*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = PixelRGB888(r, g, b);
		break;
	} }
}

// Span4ARGB()	: scan-converts a color-mapped horizontal span, featuring perspective-correct shading
//		BLENDING_OPERATOR_T			: one of RENDER_STATE_OP_*
//		TARGET_PIX_FORMAT_T			: one of RENDER_STATE_PIX_*
//		- p0,	const Vertex4ARGB &	: span 1st vertex,	input
//		- p1,	const Vertex4ARGB &	: span 2nd vertex,	input
// returns
//		nil
// note
//		- vertices must be in ascending x order

template < int BLENDING_OPERATOR_T,
		   int TARGET_PIX_FORMAT_T >

void Span4ARGB(const Vertex4ARGB &p0,
			   const Vertex4ARGB &p1)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
	const float snap_x0 = ceilf(p0.x);
	const float snap_x1 = ceilf(p1.x) - 1.f;

	if (snap_x1 < snap_x0)
		return;

	const float err_x0	= snap_x0 - p0.x;

	const float dz_dx	= rend_state.trigrad_x[0];
	const float doow_dx = rend_state.trigrad_x[1];
	const float daow_dx = rend_state.trigrad_x[2];
	const float drow_dx = rend_state.trigrad_x[3];
	const float dgow_dx = rend_state.trigrad_x[4];
	const float dbow_dx = rend_state.trigrad_x[5];

	const int span_len	= ToInt(snap_x1 - snap_x0) + 1;
	const int int_x0	= ToInt(snap_x0);

	for (int i = 0; i < span_len; i++)
	{
		const float zi	 = p0.z + (err_x0 + (float) i) * dz_dx; __unused_float(zi);
		const float oowi = p0.w + (err_x0 + (float) i) * doow_dx;
		const float aowi = p0.a + (err_x0 + (float) i) * daow_dx;
		const float rowi = p0.r + (err_x0 + (float) i) * drow_dx;
		const float gowi = p0.g + (err_x0 + (float) i) * dgow_dx;
		const float bowi = p0.b + (err_x0 + (float) i) * dbow_dx;

		const float wi = 1.f / oowi;

		const float ai = aowi * wi;
		const float ri = rowi * wi;
		const float gi = gowi * wi;
		const float bi = bowi * wi;

		switch(BLENDING_OPERATOR_T)
		{
		case RENDER_STATE_OP_COPY:

			ColorCpy< TARGET_PIX_FORMAT_T >(int_x0, i, ai, ri, gi, bi);
			break;

		case RENDER_STATE_OP_ADD:

			ColorAdd< TARGET_PIX_FORMAT_T >(int_x0, i, ai, ri, gi, bi);
			break;

		case RENDER_STATE_OP_MODULATE:

			ColorMod< TARGET_PIX_FORMAT_T >(int_x0, i, ai, ri, gi, bi);
			break;

		case RENDER_STATE_OP_DOT3:

			ColorDot< TARGET_PIX_FORMAT_T >(int_x0, i, ai, ri, gi, bi);
			break;
		}
    }
}

void Span4ARGB_cpy_ARGB8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_COPY, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4ARGB_cpy_ARGB1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_COPY, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4ARGB_cpy_RGB888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_COPY, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4ARGB_cpy_RGB565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_COPY, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4ARGB_mod_ARGB8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_MODULATE, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4ARGB_mod_ARGB1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_MODULATE, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4ARGB_mod_RGB888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_MODULATE, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4ARGB_mod_RGB565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_MODULATE, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4ARGB_dot_ARGB8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_DOT3, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4ARGB_dot_ARGB1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_DOT3, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4ARGB_dot_RGB888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_DOT3, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4ARGB_dot_RGB565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_DOT3, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4ARGB_add_ARGB8888(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_ADD, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4ARGB_add_ARGB1555(const Vertex4ARGB &p0,
							const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_ADD, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4ARGB_add_RGB888(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_ADD, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4ARGB_add_RGB565(const Vertex4ARGB &p0,
						  const Vertex4ARGB &p1)
{
	Span4ARGB< RENDER_STATE_OP_ADD, RENDER_STATE_PIX_RGB565 >(p0, p1);
}


template < int SOURCE_PIX_FORMAT_T,
		   int TARGET_PIX_FORMAT_T >

void AuxCpy(const int int_x0,
			const int i)
{
	switch(SOURCE_PIX_FORMAT_T) {

	case RENDER_STATE_PIX_ARGB1555:
	{
		const PixelARGB1555 src = *(const PixelARGB1555 *)(rend_state.aux.cursor + (int_x0 + i) * PixelARGB1555::PackedSize());

		switch(TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_ARGB1555:

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = src;
			break;

		case RENDER_STATE_PIX_ARGB8888:

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = ConvertARGB< uint32, 8, 8, 8, 8 >(src);
			break;

		case RENDER_STATE_PIX_RGB565:

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = ConvertARGB2RGB< uint32, 5, 6, 5 >(src);
			break;

		case RENDER_STATE_PIX_RGB888:

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = ConvertARGB2RGB< uint32, 8, 8, 8 >(src);
			break;
		}
		break;
	}
	case RENDER_STATE_PIX_ARGB8888:
	{
		const PixelARGB8888 src = *(const PixelARGB8888 *)(rend_state.aux.cursor + (int_x0 + i) * PixelARGB8888::PackedSize());

		switch(TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_ARGB1555:

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = ConvertARGB< uint32, 1, 5, 5, 5 >(src);
			break;

		case RENDER_STATE_PIX_ARGB8888:

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = src;
			break;

		case RENDER_STATE_PIX_RGB565:

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = ConvertARGB2RGB< uint32, 5, 6, 5 >(src);
			break;

		case RENDER_STATE_PIX_RGB888:

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = ConvertARGB2RGB< uint32, 8, 8, 8 >(src);
			break;
		}
		break;
	}
	case RENDER_STATE_PIX_RGB565:
	{
		const PixelRGB565 src = *(const PixelRGB565 *)(rend_state.aux.cursor + (int_x0 + i) * PixelRGB565::PackedSize());

		switch(TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_ARGB1555:

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = ConvertRGB2ARGB< uint32, 1, 5, 5, 5 >(src);
			break;

		case RENDER_STATE_PIX_ARGB8888:

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = ConvertRGB2ARGB< uint32, 8, 8, 8, 8 >(src);
			break;

		case RENDER_STATE_PIX_RGB565:

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = src;
			break;

		case RENDER_STATE_PIX_RGB888:

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = ConvertRGB< uint32, 8, 8, 8 >(src);
			break;
		}
		break;
	}
	case RENDER_STATE_PIX_RGB888:
	{
		const PixelRGB888 src = *(const PixelRGB888 *)(rend_state.aux.cursor + (int_x0 + i) * PixelRGB888::PackedSize());

		switch(TARGET_PIX_FORMAT_T)
		{
		case RENDER_STATE_PIX_ARGB1555:

			*(PixelARGB1555 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB1555::PackedSize()) = ConvertRGB2ARGB< uint32, 1, 5, 5, 5 >(src);
			break;

		case RENDER_STATE_PIX_ARGB8888:

			*(PixelARGB8888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelARGB8888::PackedSize()) = ConvertRGB2ARGB< uint32, 8, 8, 8, 8 >(src);
			break;

		case RENDER_STATE_PIX_RGB565:

			*(PixelRGB565 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB565::PackedSize()) = ConvertRGB< uint32, 5, 6, 5 >(src);
			break;

		case RENDER_STATE_PIX_RGB888:

			*(PixelRGB888 *)(rend_state.frame.cursor + (int_x0 + i) * PixelRGB888::PackedSize()) = src;
			break;
		}
		break;
	} }
}

// Span4_cpy()	: scan-converts a horizontal span with op COPY from an aux buffer
//		SOURCE_PIX_FORMAT_T		: one of RENDER_STATE_PIX_*
//		TARGET_PIX_FORMAT_T		: one of RENDER_STATE_PIX_*
//		- p0,	const Vertex4 &	: span 1st vertex,	input
//		- p1,	const Vertex4 &	: span 2nd vertex,	input
// returns
//		nil
template < int SOURCE_PIX_FORMAT_T,
		   int TARGET_PIX_FORMAT_T >

void Span4_cpy(const Vertex4 &p0,
			   const Vertex4 &p1)
{
	// snap left end to the next greater- or equal integer;
	// snap right end to the previous integer
    const float snap_x0 = ceilf(p0.x);
    const float snap_x1 = ceilf(p1.x) - 1.f;

    if (snap_x1 < snap_x0)
        return;

    const float err_x0	= snap_x0 - p0.x;

    const float dz_dx	= rend_state.trigrad_x[0];

	const int span_len	= ToInt(snap_x1 - snap_x0) + 1;
	const int int_x0	= ToInt(snap_x0);

    for (int i = 0; i < span_len; i++)
    {
		const float zi = p0.z + (err_x0 + (float) i) * dz_dx; __unused_float(zi);

		AuxCpy< SOURCE_PIX_FORMAT_T, TARGET_PIX_FORMAT_T >(int_x0, i);
    }
}

void Span4_cpy_ARGB8888_ARGB8888(const Vertex4 &p0,
								 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4_cpy_ARGB8888_ARGB1555(const Vertex4 &p0,
								 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4_cpy_ARGB8888_RGB888(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4_cpy_ARGB8888_RGB565(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB8888, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4_cpy_ARGB1555_ARGB8888(const Vertex4 &p0,
								 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4_cpy_ARGB1555_ARGB1555(const Vertex4 &p0,
								 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4_cpy_ARGB1555_RGB888(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4_cpy_ARGB1555_RGB565(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_ARGB1555, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4_cpy_RGB888_ARGB8888(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4_cpy_RGB888_ARGB1555(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4_cpy_RGB888_RGB888(const Vertex4 &p0,
							 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4_cpy_RGB888_RGB565(const Vertex4 &p0,
							 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB888, RENDER_STATE_PIX_RGB565 >(p0, p1);
}

void Span4_cpy_RGB565_ARGB8888(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_PIX_ARGB8888 >(p0, p1);
}

void Span4_cpy_RGB565_ARGB1555(const Vertex4 &p0,
							   const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_PIX_ARGB1555 >(p0, p1);
}

void Span4_cpy_RGB565_RGB888(const Vertex4 &p0,
							 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_PIX_RGB888 >(p0, p1);
}

void Span4_cpy_RGB565_RGB565(const Vertex4 &p0,
							 const Vertex4 &p1)
{
	Span4_cpy< RENDER_STATE_PIX_RGB565, RENDER_STATE_PIX_RGB565 >(p0, p1);
}
